import json
import os
import sys

from pdbfairy.commands import find_interactions
import tests.loader
import tests.test_find_pairs


def main(structure, max_distance, output_file):
    with open(output_file, 'w') as f:
        json.dump(tests.test_find_pairs.run_find_pairs_algorithm(
            find_pairs_brute_force, structure, max_distance), f)


def find_pairs_brute_force(structure, max_distance):
    chains = list(structure.get_chains())
    for c, chain_1 in enumerate(chains):
        for chain_2 in chains[c + 1:]:
            for atom_1 in chain_1.get_atoms():
                for atom_2 in chain_2.get_atoms():
                    if find_interactions.dist(atom_1, atom_2) <= max_distance:
                        yield (atom_1, atom_2)


if __name__ == '__main__':
    pdb_name = sys.argv[1]  # not including .pdb suffix
    try:
        max_distance = float(sys.argv[2])
    except IndexError:
        max_distance = find_interactions.MAX_DISTANCE
    loader = tests.loader.Loader(pdb_name)
    main(loader.get_structure(), max_distance, output_file=loader.output_file)



import difflib
import io

import click

from pdbfairy import utils
from pdbfairy.commands import find_interactions


@click.argument('pdb_file_1')
@click.argument('pdb_file_2')
@click.option('--max-distance', default=find_interactions.MAX_DISTANCE, help=(
    "The distance in Angstroms under which atoms should "
    "be considered to interact (default {})"
    .format(find_interactions.MAX_DISTANCE)))
def compare_interactions(pdb_file_1, pdb_file_2, max_distance):
    """
    Show how find-interactions differs for PDB_FILE_1 and PDB_FILE_2
    """
    structure_1 = utils.load_pdb_file(pdb_file_1)
    structure_2 = utils.load_pdb_file(pdb_file_2)

    with utils.capture() as (interactions_text_1, _):
        find_interactions.print_interactions(structure_1, max_distance)
    with utils.capture() as (interactions_text_2, _):
        find_interactions.print_interactions(structure_2, max_distance)

    differ = difflib.Differ()
    diff = list(differ.compare(
        interactions_text_1.getvalue().splitlines()[5:],
        interactions_text_2.getvalue().splitlines()[5:],
    ))

    print("PDB file name 1\t{}".format(structure_1.id))
    print("PDB file name 2\t{}".format(structure_2.id))
    print("Distance cutoff\t{}".format(max_distance))
    print()
    print()
    print()
    print('File\t{}'.format(diff[0].strip()))

    for line in sorted(diff[1:]):
        marker, rest = line[:2], line[2:]
        if marker == '- ':
            print('{}\t{}'.format(structure_1.id, rest))
        elif marker == '+ ':
            print('{}\t{}'.format(structure_2.id, rest))
        elif marker == '  ':
            print('both\t{}'.format(rest))
        elif marker in ('', '? '):
            pass
        else:
            raise ValueError("This should never happen: {!r}".format(marker))



import contextlib
import io
import os
import sys

from Bio import PDB


def load_pdb_file(pdb_file):
    parser = PDB.PDBParser()
    pdb_name = ''.join(os.path.splitext(os.path.basename(pdb_file))[:-1])
    return parser.get_structure(pdb_name, pdb_file)


@contextlib.contextmanager
def capture():
    """
    Thank you https://stackoverflow.com/a/10743550
    """
    oldout,olderr = sys.stdout, sys.stderr
    try:
        out=[io.StringIO(), io.StringIO()]
        sys.stdout,sys.stderr = out
        yield out
    finally:
        sys.stdout,sys.stderr = oldout, olderr
        out[0] = out[0].getvalue()
        out[1] = out[1].getvalue()



import builtins
import collections
import functools
import math
import sys

import click
import memoized
from numpy import linalg
import numpy

from pdbfairy import utils

MAX_DISTANCE = 4


@click.argument('pdb_file')
@click.option('--max-distance', default=MAX_DISTANCE, help=(
    "The distance in Angstroms under which atoms should "
    "be considered to interact (default {})".format(MAX_DISTANCE)))
def find_interactions(pdb_file, max_distance):
    """
    Find all intermolecular pairs of residues in PDB_FILE that interact

    The output will be tab-separated values, something you can
    paste into a spreadsheet (Google Spreadsheets, etc., Excel).
    """
    structure = utils.load_pdb_file(pdb_file)
    print_interactions(structure, max_distance)


def print_interactions(structure, max_distance):
    res_pairs = find_residue_pairs(structure, max_distance)

    print("PDB file name\t{}".format(structure.id))
    print("Distance cutoff\t{}".format(max_distance))
    print()
    print()
    print()
    print("Chain\tResidue number\tChain\tResidue number")

    lines = []

    def get_line(res_1, res_2):
        return (get_res_chain(res_1), get_res_number(res_1),
                get_res_chain(res_2), get_res_number(res_2))

    for res_1, res_2 in res_pairs:
        lines.append(get_line(res_1, res_2))
        lines.append(get_line(res_2, res_1))
    lines.sort()
    for line in lines:
        print('{}\t{}\t{}\t{}'.format(*line))


def find_residue_pairs(structure, max_distance):
    atom_pairs = find_pairs(structure, max_distance)
    res_pairs = set()
    for atom_1, atom_2 in atom_pairs:
        res_pairs.add((atom_1.parent, atom_2.parent))
    return res_pairs


def find_pairs(structure, max_distance):
    atoms = list(structure.get_atoms())
    atom_store = AtomStore(atoms, max_distance=max_distance)
    for atom in atoms:
        yield from atom_store.find_neighbors(atom)
        atom_store.remove(atom)


class AtomStore(object):
    def __init__(self, atoms, max_distance):
        # max distance for two atoms to be considered "neighbors"
        self.max_distance = max_distance
        self._atoms_by_cube = collections.defaultdict(set)
        self._atoms_by_chain = collections.defaultdict(set)
        for atom in atoms:
            self.add(atom)

    def add(self, atom):
        self._atoms_by_cube[tuple(self._get_cube(atom))].add(atom)
        self._atoms_by_chain[self._get_chain(atom)].add(atom)

    def remove(self, atom):
        self._atoms_by_cube[tuple(self._get_cube(atom))].remove(atom)
        self._atoms_by_chain[self._get_chain(atom)].remove(atom)

    def find_neighbors(self, atom):
        cube = self._get_cube(atom)
        atoms_in_chain = self._atoms_by_chain[self._get_chain(atom)]
        yield from (
            (atom, a)
            for offset in self._offsets
            for a in self._atoms_by_cube[tuple(cube + offset)] - atoms_in_chain
            if dist(a, atom) <= self.max_distance
        )

    @memoized.memoized
    def _get_cube(self, atom):
        return numpy.floor(atom.coord / self.max_distance)

    @memoized.memoized
    def _get_chain(self, atom):
        return atom.parent.parent

    _offsets = [
        numpy.array([i, j, k])
        for i in [-1, 0, 1]
        for j in [-1, 0, 1]
        for k in [-1, 0, 1]
    ]


def dist(atom_1, atom_2):
    return linalg.norm(atom_1.coord - atom_2.coord)


def get_res_chain(res):
    return res.parent.id


def get_res_number(res):
    return res.id[1]



import click

from pdbfairy.commands import find_interactions, compare_interactions


@click.group()
def main():
    pass


main.command()(find_interactions.find_interactions)
main.command()(compare_interactions.compare_interactions)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pdbfairy', 'pdbfairy.commands']

package_data = \
{'': ['*']}

install_requires = \
['biopython>=1.74,<2.0',
 'click>=7.0,<8.0',
 'memoized>=0.3.0,<0.4.0',
 'numpy>=1.16,<1.17']

entry_points = \
{'console_scripts': ['pdbfairy = pdbfairy.main:main']}

setup_kwargs = {
    'name': 'pdbfairy',
    'version': '0.1.2',
    'description': '',
    'long_description': None,
    'author': 'Danny Roberts',
    'author_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.4,<4.0',
}


setup(**setup_kwargs)
